@@ -17,8 +17,9 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "mbexp.h"
+#define _GNU_SOURCE
 
+#include "mbexp.h"
 
 void /* Not Used  */
 mb_init(void)
@@ -106,7 +107,8 @@ mb_col_set (MBColor *col, char *spec)
 	    }
 	  else
 	    {
-	      fprintf(stderr, "mbcolor: failed to parse color %s\n", spec);
+	      if (mb_want_warnings())
+		fprintf(stderr, "mbcolor: failed to parse color %s\n", spec);
 	      return False;
 	    } 
 
@@ -121,7 +123,8 @@ mb_col_set (MBColor *col, char *spec)
 			   DefaultColormap(pb->dpy, pb->scr),
 			   spec, &col->xcol))
 	    {
-	      fprintf(stderr, "mbcolor: failed to parse color %s\n", spec);
+	      if (mb_want_warnings())
+		fprintf(stderr, "mbcolor: failed to parse color %s\n", spec);
 	      return False;
 	    } 
 
@@ -345,7 +348,7 @@ _mb_font_load(MBFont *font)
 			    XFT_SIZE, XftTypeDouble   , (double)font->pt_size,
 			    XFT_WEIGHT, XftTypeInteger, weight,
 			    XFT_SLANT, XftTypeInteger , slant,
-			    0);
+			    NULL);
 
   if (font->font != NULL ) result = 2;
 
@@ -373,11 +376,13 @@ _mb_font_load(MBFont *font)
     { 
       result = 1;  /* XXX: should retry with weight, slant set to '*' */ 
 
-      fprintf(stderr, "mbfont: failed to load %s, falling back to fixed\n", font_spec);
+      if (mb_want_warnings())
+	fprintf(stderr, "mbfont: failed to load %s, falling back to fixed\n", font_spec);
 
       if ((font->font = XLoadQueryFont(font->dpy, "fixed")) == NULL)
 	{
-	  fprintf(stderr, "mbfont: fixed failed, no usable fonts\n");
+	  if (mb_want_warnings())
+	    fprintf(stderr, "mbfont: fixed failed, no usable fonts\n");
 	  result = 0;
 	}
     }
@@ -405,6 +410,10 @@ mb_font_new(Display *dpy,
 #endif
   
   font = malloc(sizeof(MBFont));
+
+  if (font == NULL)
+    return NULL;
+
   memset(font, 0, sizeof(MBFont));
 
   if (family != NULL)
@@ -424,9 +433,12 @@ mb_font_new(Display *dpy,
    font->pgo_fontmap = pango_xft_get_font_map (font->dpy, DefaultScreen(dpy));
    font->fontdes     = pango_font_description_new ();
 
-   /* -- Needed ?
-   pango_context_set_language (w->pgo, pango_language_from_string ("ar_AE"));
-   */
+   /* If Pango is mis-setup the above will fail */
+   if (font->pgo_context == NULL || font->pgo_fontmap == NULL || font->fontdes == NULL)
+     {
+       free(font);
+       return NULL;
+     }
 
 #elif defined (USE_XFT)
 
@@ -580,8 +592,11 @@ MBFont*
 mb_font_new_from_string(Display *dpy, char *spec) 
 {
   MBFont *font = mb_font_new(dpy, NULL);
-  mb_font_set_from_string(font, spec); 
-  return font;
+
+  if (font)
+    return mb_font_set_from_string(font, spec); 
+
+  return NULL;
 }
 
 MBFont*
@@ -776,7 +791,7 @@ mb_font_get_txt_width(MBFont        *font,
    */
 
   attr_list = pango_attr_list_new (); /* no markup - empty attributes */
-  str       = strdup(txt);
+  str       = strdup((char*)txt);
 
   str[byte_len] = '\0';
   /*
@@ -805,8 +820,8 @@ mb_font_get_txt_width(MBFont        *font,
 
        pango_glyph_string_extents (  glyphs,
 				     this->analysis.font,
-				     &rect,
-				     NULL);
+				     NULL,
+				     &rect);
 
        width += (( rect.x + rect.width ) / PANGO_SCALE);
        
@@ -893,11 +908,130 @@ _clip_some_text (MBFont         *font,
 		 int             encoding,
 		 int             opts)
 {
-  int len = strlen(txt);
+#ifdef USE_PANGO
+  /* We cache the width of the elipsis for the given font */
+  static int     trailing_width = 0;
+  static MBFont *font_prev = NULL;
+  
+  GList         *items     = NULL, *items_head = NULL;
+  PangoAttrList *attr_list = NULL;
+  int            width     = 0;
+  int            length    = 0;
+  int            done = 0;
+  int            bytelen = strlen((char*)txt);
+
+  /* we can't clip single char string */
+  if (bytelen < 2) return 0;
+
+  if (font != font_prev || !_mb_font_is_font_object_fresh (font))
+    trailing_width = 0; /* different font, recalculate */
+  
+  if (!_mb_font_is_font_object_fresh (font)) /* XXX Make define ? */
+    _mb_font_load(font);
+
+  attr_list = pango_attr_list_new (); /* no markup - empty attributes */
+  
+  /* Measure the trailing elipsis */
+  if ((opts & MB_FONT_RENDER_OPTS_CLIP_TRAIL) && trailing_width == 0)
+    {
+      items = pango_itemize (font->pgo_context, ".", 0, 1, attr_list, NULL);
+
+      if (items)
+	{
+	  PangoItem        *this   = (PangoItem *)items->data;
+	  PangoGlyphString *glyphs = pango_glyph_string_new ();
+	  PangoRectangle    rect;
+       
+	  pango_shape  (".", 1, &this->analysis, glyphs);
+
+	  pango_glyph_string_extents (  glyphs,
+					this->analysis.font,
+					NULL,
+					&rect);
+
+	  trailing_width = ((3 * ( rect.x + rect.width )) / PANGO_SCALE);
+       
+	  pango_item_free (this);
+	  pango_glyph_string_free (glyphs);
+
+	  g_list_free (items);
+
+	  font_prev = font;
+	}
+    }
+
+  /* Measure the actual string */
+  items_head = items = pango_itemize (font->pgo_context, (char*)txt, 0, 
+				      strlen ((char*)txt), attr_list, NULL);
+
+   while (items)
+     {
+       PangoItem *this = (PangoItem *)items->data;
+
+       if (!done)
+	 {
+	   PangoGlyphString *glyphs = pango_glyph_string_new ();
+	   PangoRectangle    rect;
+	   int               rtl = (this->analysis.level & 1);
+	   
+	   pango_shape  ((char*)&txt[this->offset], this->length, 
+			 &this->analysis, glyphs);
+
+	   pango_glyph_string_extents (  glyphs,
+					 this->analysis.font,
+					 NULL,
+					 &rect);
+
+	   width += (( rect.x + rect.width ) / PANGO_SCALE);
+	   length += this->length;
+       
+	   if (width + trailing_width == max_width)
+	     {
+	       done = 1;
+	     }
+	   else if (width + trailing_width > max_width)
+	     {
+	       /* The required break is somewhere in this item */
+	       int k;
+	       int trim = width + trailing_width - max_width;
+	       int byte_off = 0;
+	       
+	       for (k = 0; k < glyphs->num_glyphs; ++k)
+		 {
+		   int indx = rtl ? k : glyphs->num_glyphs - k - 1;
+		   
+		   trim -= (glyphs->glyphs[indx].geometry.x_offset +
+			    glyphs->glyphs[indx].geometry.width) / PANGO_SCALE;
+
+		   byte_off = glyphs->log_clusters[indx];
+		   
+		   if (trim <= 0)
+		     break;
+		 }
+
+	       length -= (this->length - byte_off);
+	       done = 1;
+	     }
+	   
+	   pango_glyph_string_free (glyphs);
+	 }
+       
+       pango_item_free (this);
+       items = items->next;
+     }
+
+   
+   pango_attr_list_unref (attr_list);
+   
+   if (items_head) g_list_free (items_head);
+
+   return length;
+   
+#else
+  int len = strlen((char*)txt);
 
   /* we cant clip single char string */
   if (len < 2) return 0;
-
   /* XXX RTL case for pango */
 
   if (opts & MB_FONT_RENDER_OPTS_CLIP_TRAIL)
@@ -905,8 +1039,7 @@ _clip_some_text (MBFont         *font,
       unsigned char *str = malloc(len+5);
       memset(str, 0, len+5);
 
-      /* len += 2; */
-      strcpy(str, txt);
+      strcpy((char*)str, (char*)txt);
 
       do {
 	  /* go back a glyth */
@@ -944,6 +1077,7 @@ _clip_some_text (MBFont         *font,
       else len--;
     }
   return len;
+#endif
 }
 
 void
@@ -974,10 +1108,10 @@ _render_some_text (MBFont        *font,
    */
 
    attr_list = pango_attr_list_new (); /* no markup - empty attributes */
-   str       = strdup(text);
+   str       = (unsigned char*)strdup((char*)text);
 
    /* analyse string, breaking up into items */
-   items_head = items = pango_itemize (font->pgo_context, str, 
+   items_head = items = pango_itemize (font->pgo_context, (char*)str, 
 				       0, bytes_to_render,
 				       attr_list, NULL);
 
@@ -988,7 +1122,7 @@ _render_some_text (MBFont        *font,
        PangoRectangle    rect;
        
        /* shape current item into run of glyphs */
-       pango_shape  (&str[this->offset], this->length, 
+       pango_shape  ((char*)&str[this->offset], this->length, 
 		     &this->analysis, glyphs);
 
        /* render the glyphs */
@@ -1001,8 +1135,8 @@ _render_some_text (MBFont        *font,
        /* calculate rendered area */
        pango_glyph_string_extents (glyphs,
 				   this->analysis.font,
-				   &rect,
-				   NULL);
+				   NULL,
+				   &rect);
 
        x += ( rect.x + rect.width ) / PANGO_SCALE; /* XXX Correct ? */
        
@@ -1049,7 +1183,7 @@ mb_font_render_simple (MBFont          *font,
 		       int              encoding,
 		       MBFontRenderOpts opts )
 {
-  int render_w = 0, len = 0, orig_len = 0;
+  int render_w = 0, len = 0;
   unsigned char *str = NULL;
   Bool want_dots = False;
 
@@ -1057,20 +1191,21 @@ mb_font_render_simple (MBFont          *font,
 
   if (font->col == NULL) 
     {
-      fprintf(stderr, 
-	      "libmb: **error** font has no color set. unable to render\n");
+      if (mb_want_warnings())
+	fprintf(stderr, 
+		"libmb: **error** font has no color set. unable to render\n");
       return 0;
     }
 
   if (!_mb_font_is_font_object_fresh (font))
     _mb_font_load(font);
 
-  orig_len = len = strlen(text);
+  len = strlen((char*)text);
 
   str = malloc(len+3);
   memset(str, 0, len+3);
 
-  strcpy(str, text);
+  strcpy((char*)str, (char*)text);
 
   render_w = mb_font_get_txt_width(font, str, len, encoding);
 
@@ -1084,13 +1219,18 @@ mb_font_render_simple (MBFont          *font,
        */
 
 
-
       len = _clip_some_text (font, width, str, encoding, opts);
 
       if (!len) { free(str); return 0; }
       
       if ((opts & MB_FONT_RENDER_OPTS_CLIP_TRAIL) && len > 3)
+	{
+	  /* Avoid having a space before the elipsis */
+	  while (len-1 >= 0 && str[len-1] == ' ')
+	    len--;
+
 	  want_dots = True;
+	}
     }
   else
     {
@@ -1137,7 +1277,7 @@ mb_font_render_simple_get_width (MBFont          *font,
 				 int              encoding,
 				 MBFontRenderOpts opts )
 {
-  int render_w = 0, len = 0, orig_len = 0;
+  int render_w = 0, len = 0;
   unsigned char *str = NULL;
   Bool want_dots = False;
 
@@ -1146,12 +1286,12 @@ mb_font_render_simple_get_width (MBFont          *font,
   if (!_mb_font_is_font_object_fresh (font))
     _mb_font_load(font);
 
-  orig_len = len = strlen(text);
+  len = strlen((char*)text);
 
   str = malloc(len+3);
   memset(str, 0, len+3);
 
-  strcpy(str, text);
+  strcpy((char*)str, (char*)text);
 
   render_w = mb_font_get_txt_width(font, str, len, encoding);
 
@@ -1256,7 +1396,7 @@ mb_layout_set_text(MBLayout       *layout,
 {
   if (layout->txt) free(layout->txt);
 
-  layout->txt = strdup(text);
+  layout->txt = (unsigned char*)strdup((char*)text);
   layout->txt_encoding = encoding;
 }
 
@@ -1303,7 +1443,7 @@ _mb_layout_render_magic (MBLayout        *layout,
 			 MBFontRenderOpts opts,
 			 Bool             do_render)
 {
-  unsigned char *orig_p, *p = strdup(layout->txt);
+  unsigned char *orig_p, *p = (unsigned char*)strdup((char*)layout->txt);
   unsigned char *q = p;
   unsigned char *backtrack = NULL;
   int            v_offset  = 0;
@@ -1325,7 +1465,7 @@ _mb_layout_render_magic (MBLayout        *layout,
 	  
 	  /* XXX q should be current_line_start */
 
-	  cur_width = mb_font_get_txt_width(layout->font, q, strlen(q), 
+	  cur_width = mb_font_get_txt_width(layout->font, q, strlen((char*)q), 
 					    layout->txt_encoding) ;
 	  
 	  if (cur_width > layout->width )
@@ -1404,7 +1544,7 @@ mb_layout_render (MBLayout        *layout,
 
   if (layout->_have_autocalc_size) /* Easy case */
     {
-      unsigned char *str = strdup(layout->txt), *start = NULL, *orig = NULL;
+      char *str = strdup((char*)layout->txt), *start = NULL, *orig = NULL;
 
       orig = str;
 
@@ -1425,7 +1565,7 @@ mb_layout_render (MBLayout        *layout,
 		                 x,
 		                 y,
 				 layout->width,
-				 start,
+				 (unsigned char*)start,
 		                 layout->txt_encoding,
 		                 0 );
 	  
@@ -1532,8 +1672,8 @@ mb_util_utf8_len()
 int 				/* XXX this can be much smaller see glib */
 mb_util_next_utf8_char(unsigned char **string)
 {
-  unsigned char *s, mask;
-  int length;
+  unsigned char *s;
+  int length, tmp;
   
   s = *string;
 
@@ -1546,25 +1686,26 @@ mb_util_next_utf8_char(unsigned char **string)
   }
   
   if((*s & 0xe0) == 0xc0) { /* 110xxxxx followed by one trailer */
-    mask = 0x1f; length = 1;
+    length = 1;
   } else if ((*s & 0xf0) == 0xe0) { 
     /* 1110xxxx followed by two trailers */     
-    mask = 0x0f; length = 2;
+    length = 2;
   } else if ((*s & 0xf8) == 0xf0) { 
     /* 11110xxx followed by three trailers */     
-    mask = 0x07; length = 3;
+    length = 3;
   } else if((*s & 0xfc) == 0xf8) { 
     /* 111110xx followed by four trailers */     
-    mask = 0x03; length = 4;
+    length = 4;
   } else if((*s * 0xfe) == 0xfc) { 
     /* 1111110x followed by five trailers */     
-    mask = 0x01; length = 5;
+    length = 5;
   } else { /* 1111111x is invalid in UTF8 */     
     return -1;
   }
 	    
   *s++;
-  while(length-- > 0) {
+  tmp = length;
+  while(tmp-- > 0) {
     if((*s & 0xc0) != 0x80) { /* trailer must be 10xxxxxx */
       /* ERROR */ 
       return -1;
@@ -1573,6 +1714,6 @@ mb_util_next_utf8_char(unsigned char **string)
   } 
   
   *string = s; 
-  return length;
+  return length + 1;
 
 }
