@@ -22,6 +22,7 @@
 
 #include <strings.h>
 #include "mbpixbuf.h"
+#include "mbutil.h"
 
 #define BYTE_ORD_24_RGB  0
 #define BYTE_ORD_24_RBG  1
@@ -150,7 +151,8 @@ _load_jpg_file( const char *file,
   unsigned char *data = NULL;
 
   if ((infile = fopen(file, "rb")) == NULL) {
-    fprintf(stderr, "mbpixbuf: can't open %s\n", file);
+    if (mb_want_warnings())
+      fprintf(stderr, "mbpixbuf: can't open %s\n", file);
     return NULL;
   }
 
@@ -176,8 +178,9 @@ _load_jpg_file( const char *file,
 
   if( cinfo.output_components != 3 ) 
     {
-      fprintf( stderr, "mbpixbuf: jpegs with %d channles not supported\n", 
-	       cinfo.output_components );
+      if (mb_want_warnings())
+	fprintf( stderr, "mbpixbuf: jpegs with %d channles not supported\n", 
+		 cinfo.output_components );
       jpeg_finish_decompress(&cinfo);
       jpeg_destroy_decompress(&cinfo);
       return NULL;
@@ -222,6 +225,7 @@ _load_png_file( const char *file,
   png_structp png_ptr;
   png_infop info_ptr;
   png_bytep *row_pointers;
+  png_byte ct;
 
   if ((fd = fopen( file, "rb" )) == NULL) return NULL;
 
@@ -245,7 +249,7 @@ _load_png_file( const char *file,
     return NULL;
   }
 
-  if ( setjmp( png_ptr->jmpbuf ) ) {
+  if ( setjmp( png_jmpbuf( png_ptr ) ) ) {
     png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
     fclose(fd);
     return NULL;
@@ -266,9 +270,11 @@ _load_png_file( const char *file,
   if (( color_type == PNG_COLOR_TYPE_GRAY )||
       ( color_type == PNG_COLOR_TYPE_GRAY_ALPHA ))
     png_set_gray_to_rgb(png_ptr);
- 
-  if ( info_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA 
-       || info_ptr->color_type == PNG_COLOR_TYPE_GRAY_ALPHA
+
+  ct = png_get_color_type(png_ptr, info_ptr);
+
+  if ( ct == PNG_COLOR_TYPE_RGB_ALPHA
+       || ct == PNG_COLOR_TYPE_GRAY_ALPHA
        )
     *has_alpha = 1;
   else
@@ -374,7 +380,8 @@ _load_xpm_file( MBPixbuf *pb, const char *filename, int *w, int *h, int *has_alp
 		  sscanf(line, "%i %i %i %i", w, h, &ncolors, &cpp);
                   if (ncolors > 32766 || cpp > 5 || *w > 32767 || *h > 32767)
 		    {
-		      fprintf(stderr, "xpm file invalid");
+		      if (mb_want_warnings())
+			fprintf(stderr, "xpm file invalid");
 		      fclose(file);
 		      free(line);
 		      fclose(file);
@@ -418,7 +425,7 @@ _load_xpm_file( MBPixbuf *pb, const char *filename, int *w, int *h, int *has_alp
 		      col[0] = 0;
 		      s[0] = 0;
 		      len = strlen(line);
-		      strncpy(cmap[j].str, line, cpp);
+		      strncpy((char*)cmap[j].str, line, cpp);
 		      cmap[j].str[cpp] = 0;
 		      cmap[j].r = -1;
 		      cmap[j].transp = 0;
@@ -537,7 +544,7 @@ _load_xpm_file( MBPixbuf *pb, const char *filename, int *w, int *h, int *has_alp
 			  i--;
 			  for (j = 0; j < ncolors; j++)
 			    {
-			      if (!strcmp(col, cmap[j].str))
+			      if (!strcmp(col, (char*)cmap[j].str))
 				{
 				  if (transp && cmap[j].transp)
 				    {
@@ -643,7 +650,8 @@ _paletteAlloc(MBPixbuf *pb)
 	  
 	  if (!XAllocColor(pb->dpy, pb->root_cmap, &xcl))
 	    {
-	      //printf("alloc color failed\n");
+	      if (mb_want_warnings())
+		printf("alloc color failed\n");
 	    }
 	  is_used = 0;
 	  for (j = 0; j < num_used; j++)
@@ -678,7 +686,8 @@ _paletteAlloc(MBPixbuf *pb)
 	  
 	  if (!XAllocColor(pb->dpy, pb->root_cmap, &xcl))
 	    {
-	      fprintf(stderr, "libmb: alloc color failed\n");
+	      if (mb_want_warnings())
+		fprintf(stderr, "libmb: alloc color failed\n");
 	    }
 	  is_used = 0;
 	  for (j = 0; j < num_used; j++)
@@ -716,7 +725,13 @@ mb_pixbuf_get_pixel(MBPixbuf *pb, int r, int g, int b, int a)
 	case 15:
 	  return ((r & 0xf8) << 7) | ((g & 0xf8) << 2) | ((b & 0xf8) >> 3);
 	case 16:
-	  return ((r & 0xf8) << 8) | ((g & 0xfc) << 3) | ((b & 0xf8) >> 3);
+	  switch (pb->byte_order)
+	    {
+	    case BYTE_ORD_24_RGB:
+	      return ((r & 0xf8) << 8) | ((g & 0xfc) << 3) | ((b & 0xf8) >> 3);
+	    case BYTE_ORD_24_BGR:
+	      return ((b & 0xf8) << 8) | ((g & 0xfc) << 3) | ((r & 0xf8) >> 3);
+	    }
 	case 24:
 	case 32:
 	  switch (pb->byte_order)
@@ -907,6 +922,7 @@ mb_pixbuf_new_extended(Display *dpy,
 	  fprintf(stderr, "mbpixbuf: unable to use XShm. DISPLAY remote?\n");
 	  pb->have_shm = False;
 	}
+      else XShmDetach(pb->dpy, &shminfo);
 
       shmdt(shminfo.shmaddr);
       shmctl(shminfo.shmid, IPC_RMID, 0);
@@ -1013,6 +1029,54 @@ mb_pixbuf_img_new_from_int_data(MBPixbuf            *pixbuf,
   return img;
 }
 
+MBPixbufImage *
+mb_pixbuf_img_new_from_long_data(MBPixbuf            *pixbuf,
+				const unsigned long  *data,
+				int                  width,
+				int                  height)
+{
+  MBPixbufImage *img;
+  int            i=0,x,y;
+
+  img = mb_pixbuf_img_rgba_new(pixbuf, width, height);
+
+  if (pixbuf->internal_bytespp == 3)
+    {
+      unsigned char *p = img->rgba;
+
+      for (y=0; y<height; y++)
+	for (x=0; x<width; x++)
+	  {
+	    *p++ = (data[i] >> 16) & 0xff;
+	    *p++ = (data[i] >> 8) & 0xff;
+	    *p++ = data[i] & 0xff;
+	    *p++ = data[i] >> 24;
+	    i++;
+	  }
+    }
+  else
+    {
+      unsigned char *p = img->rgba, r,g,b,a;
+
+      for (y=0; y<height; y++)
+	for (x=0; x<width; x++)
+	  {
+	    r = ((data[i] >> 16) & 0xff);
+	    g = ((data[i] >> 8) & 0xff);
+	    b = (data[i] & 0xff);
+	    a = (data[i] >> 24);
+
+	    internal_rgb_to_16bpp_pixel(r,g,b,p);
+	    internal_16bpp_pixel_next(p);
+	    *p++ = a;
+
+	    i++;
+	  }
+    }
+
+  return img;
+}
+
 MBPixbufImage *
 mb_pixbuf_img_new_from_data(MBPixbuf            *pixbuf, 
 			    const unsigned char *data,
@@ -1086,7 +1150,8 @@ mb_pixbuf_img_new_from_x_drawable (MBPixbuf *pb,
   int num_of_cols = 1 << pb->depth;
 
   Window chld;
-  unsigned int rx, rw, rh, rb, rdepth;
+  int          rx;
+  unsigned int rw, rh, rb, rdepth;
 
   XShmSegmentInfo shminfo; 
 
@@ -1389,8 +1454,7 @@ mb_pixbuf_img_fill(MBPixbuf *pb,
 	  {
 	    internal_rgb_to_16bpp_pixel(r,g,b,p);
 	    internal_16bpp_pixel_next(p);
-	    if (img->has_alpha) 
-	      { *p = a; *p++; }
+	    if (img->has_alpha) *p++ = a;
 	  }
     }
   else
@@ -1849,7 +1913,7 @@ mb_pixbuf_img_render_to_drawable_with_gc(MBPixbuf    *pb,
 
 	  if (img->ximg->data == (char *)-1)
 	    {
-	      fprintf(stderr, "MBPIXBUF ERROR: SHM can't attach SHM Segment for Shared XImage, falling back to XImages\n");
+	      fprintf(stderr, "mbpixbuf: SHM can't attach SHM Segment for Shared XImage, falling back to XImages\n");
 	      XDestroyImage(img->ximg);
 	      shmctl(shminfo.shmid, IPC_RMID, 0);
 	    }
@@ -1879,12 +1943,11 @@ mb_pixbuf_img_render_to_drawable_with_gc(MBPixbuf    *pb,
 	  for(y=0; y<img->height; y++)
 	    for(x=0; x<img->width; x++)
 	      {
-		/* Below is potentially dangerous.  
-		 */
-		pixel =  ( *p | (*(p+1) << 8)); 
+		internal_16bpp_pixel_to_rgb(p, r, g, b);
+		internal_16bpp_pixel_next(p);
+		a = ((img->has_alpha) ?  *p++ : 0xff);
 
-		p +=  ((img->has_alpha) ?  3 : 2);
-		
+		pixel = mb_pixbuf_get_pixel(pb, r, g, b, a);
 		XPutPixel(img->ximg, x, y, pixel);
 	      }
 	}
@@ -1957,7 +2020,8 @@ mb_pixbuf_img_render_to_mask(MBPixbuf    *pb,
 
 	  if (img->ximg->data == (char *)-1)
 	    {
-	      fprintf(stderr, "MBPIXBUF ERROR: SHM can't attach SHM Segment for Shared XImage, falling back to XImages\n");
+	      if (mb_want_warnings())
+		fprintf(stderr, "mbpixbuf: SHM can't attach SHM Segment for Shared XImage, falling back to XImages\n");
 	      XDestroyImage(img->ximg);
 	      shmctl(shminfo.shmid, IPC_RMID, 0);
 	    }
